SystemOrganization addCategory: #'System-Undo'!

Object subclass: #AbstractUndoRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Undo'!

!AbstractUndoRecord commentStamp: '<historical>' prior: 0!
This class represents the abstract structure of records used in UndoManager!

!AbstractUndoRecord methodsFor: 'accessing' stamp: 'dg 6/21/2007 16:00'!
description
	"return a string describing this record, e.g. for use in history-representation"
	^self subclassResponsibility! !

!AbstractUndoRecord methodsFor: 'command' stamp: 'dg 6/4/2007 18:36'!
do
	self subclassResponsibility! !

!AbstractUndoRecord methodsFor: 'command' stamp: 'dg 6/11/2007 16:13'!
redo
	"by default, do the #do again"
	self do! !

!AbstractUndoRecord methodsFor: 'command' stamp: 'dg 6/4/2007 18:36'!
undo
	self subclassResponsibility! !

!AbstractUndoRecord methodsFor: 'accessing' stamp: 'dg 6/21/2007 16:00'!
undoable
	"Return true if this record is undoable. Return false if it is not and destroys the previous history"
	^self subclassResponsibility! !

AbstractUndoRecord subclass: #UndoRecord
	instanceVariableNames: 'undoable doMessage undoMessage description'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Undo'!

!UndoRecord classSide methodsFor: 'instance creation' stamp: 'dg 6/5/2007 11:49'!
do: doMessageSend undo: undoMessageSend

	| anUndoRecord |
	anUndoRecord := self new.
	anUndoRecord doMessage: doMessageSend.
	undoMessageSend 
		ifNotNil: [ 
			anUndoRecord undoMessage: undoMessageSend.
			anUndoRecord undoable: true.]
		ifNil: [anUndoRecord undoable: false].
	^anUndoRecord.! !

!UndoRecord classSide methodsFor: 'instance creation' stamp: 'dg 6/5/2007 11:47'!
doArray: doArray undoArray: undoArray
" array: #(receiver selector arguments) "
	| anUndoRecord |
	anUndoRecord := self new.
	anUndoRecord doMessage: (SophieMessageSend 
		receiver: 	(doArray first) 
		selector: 	(doArray second) 
		arguments: 	(doArray third)).
	undoArray 
		ifNotNil: [ 
			anUndoRecord undoMessage: (SophieMessageSend 
				receiver: 	(undoArray first) 
				selector: 	(undoArray second) 
				arguments: 	(undoArray third)).
			anUndoRecord undoable: true.]
		ifNil: [anUndoRecord undoable: false].
	^anUndoRecord.! !

!UndoRecord methodsFor: 'accessing' stamp: 'dg 6/21/2007 15:40'!
description
	^description! !

!UndoRecord methodsFor: 'accessing' stamp: 'dg 6/21/2007 15:40'!
description: aString
	description := aString! !

!UndoRecord methodsFor: 'command' stamp: 'dg 6/5/2007 11:26'!
do
	^doMessage value.! !

!UndoRecord methodsFor: 'accessing' stamp: 'dg 6/5/2007 11:22'!
doMessage
	"Answer the doMessage of the receiver"
	^doMessage! !

!UndoRecord methodsFor: 'accessing' stamp: 'dg 6/5/2007 11:22'!
doMessage: aValue
	"Modify the receiver's doMessage"
	doMessage := aValue! !

!UndoRecord methodsFor: 'printing' stamp: 'dg 6/21/2007 15:39'!
printOn: aStream
	super printOn: aStream.
	aStream space; nextPutAll: self description printString.! !

!UndoRecord methodsFor: 'command' stamp: 'dg 6/5/2007 11:28'!
undo
	^undoMessage
		ifNotNil: [undoMessage value].! !

!UndoRecord methodsFor: 'accessing' stamp: 'dg 6/5/2007 11:22'!
undoMessage
	"Answer the undoMessage of the receiver"
	^undoMessage! !

!UndoRecord methodsFor: 'accessing' stamp: 'dg 6/5/2007 11:22'!
undoMessage: aValue
	"Modify the receiver's undoMessage"
	undoMessage := aValue! !

!UndoRecord methodsFor: 'accessing' stamp: 'dg 6/5/2007 11:53'!
undoable
	^undoable! !

!UndoRecord methodsFor: 'accessing' stamp: 'dg 6/5/2007 11:53'!
undoable: aBoolean
	undoable := aBoolean! !

AbstractUndoRecord subclass: #UndoRecordGroup
	instanceVariableNames: 'description subRecords'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Undo'!

!UndoRecordGroup methodsFor: 'accessing' stamp: 'dg 6/5/2007 11:34'!
add: aSophieUndoEntry

	subRecords addLast: aSophieUndoEntry! !

!UndoRecordGroup methodsFor: 'accessing' stamp: 'dg 6/21/2007 16:13'!
description
	^description! !

!UndoRecordGroup methodsFor: 'accessing' stamp: 'dg 6/21/2007 16:14'!
description: aString
	description := aString! !

!UndoRecordGroup methodsFor: 'command' stamp: 'dg 6/5/2007 11:31'!
do

	subRecords do: [ :each |
		each do].! !

!UndoRecordGroup methodsFor: 'initialize' stamp: 'dg 6/4/2007 18:40'!
initialize
	super initialize.
	subRecords := OrderedCollection new.! !

!UndoRecordGroup methodsFor: 'testing' stamp: 'swr 10/24/2007 11:34'!
isEmpty
	"Return true if we have no subrecords"
	^subRecords isEmpty! !

!UndoRecordGroup methodsFor: 'printing' stamp: 'dg 6/21/2007 15:39'!
printOn: aStream
	super printOn: aStream.
	aStream space; nextPutAll: self description printString.! !

!UndoRecordGroup methodsFor: 'command' stamp: 'dg 6/12/2007 17:41'!
redo

	subRecords do: [ :each |
		each redo].! !

!UndoRecordGroup methodsFor: 'command' stamp: 'dg 6/5/2007 11:30'!
undo

	subRecords reverseDo: [ :each |
		each undo].! !

!UndoRecordGroup methodsFor: 'accessing' stamp: 'dg 6/5/2007 11:55'!
undoable

	subRecords do: [ :each |
		each undoable ifFalse: [ ^false]
	].

	^true! !

Object subclass: #UndoManager
	instanceVariableNames: 'undoHistory redoHistory groupStack historyInvalid maxHistoryLength'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Undo'!

!UndoManager commentStamp: '<historical>' prior: 0!
UndoManager manages execution and storing of UndoRecords through #doAndAddRecord:.
It allows hierarchical grouping of UndoRecords through #openRecordGroup, #openRecordGroupNamed: and #closeRecordGroup.
Undo and redo is invoked by #undo and #redo.

See UndoManagerTest for someExamples.

Structure:
 undoHistory		OrderedCollection -- holds UndoRecords for undo
 redoHistory			OrderedCollection -- holds UndoRecords for redo
 maxHistoryLength	Integer -- Old UndoRecords in undoHistory are removed to keep it within given limits (0 for no Limit)
 groupStack			OrderedCollection -- temporary dataStructure to manage hierarchical grouping of UndoRecords
					by #openRecordGroup and closeRecordGroup
 historyInvalid		Boolean -- flag to indicate, that the history is invalid and it must be deleted including the groupStack
!

UndoManager subclass: #LoggingUndoManager
	instanceVariableNames: 'logStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Undo'!

!LoggingUndoManager commentStamp: '<historical>' prior: 0!
This class extends the UndoManager with logging for debugging and bugtracking.

#log returns a ByteString representing the current log.!

!LoggingUndoManager methodsFor: 'transaction' stamp: 'dg 7/6/2007 18:03'!
closeRecordGroup

	super closeRecordGroup.

	logStream cr.
	self groupStack size timesRepeat: [logStream tab].
	logStream nextPutAll: 'closeRecordGroup'.
! !

!LoggingUndoManager methodsFor: 'transaction' stamp: 'JMM 7/6/2007 12:48'!
doAndAddRecord: undoRecord
	| results |

	results := super doAndAddRecord: undoRecord.

	logStream cr.
	self groupStack size timesRepeat: [logStream tab].
	logStream nextPutAll: 'doAndAddRecord: '; print: undoRecord.
	^results
! !

!LoggingUndoManager methodsFor: 'initialize' stamp: 'dg 7/6/2007 18:21'!
initLogStreamLimited: aNumber

	(logStream := LimitedWriteStream on: (String new: 5))
		setLimit: aNumber
		limitBlock: [
			logStream originalContents allButFirst.
			logStream reset]! !

!LoggingUndoManager methodsFor: 'initialize' stamp: 'JMM 12/16/2007 11:43'!
initialize

	super initialize.

	self initLogStreamLimited: 10000.! !

!LoggingUndoManager methodsFor: 'accessing' stamp: 'dg 7/6/2007 17:58'!
log

	^(logStream originalContents copyFrom: logStream position + 1 to: logStream size), (logStream originalContents copyFrom: 1 to: logStream position)! !

!LoggingUndoManager methodsFor: 'transaction' stamp: 'dg 7/6/2007 17:47'!
openRecordGroup

	logStream cr.
	self groupStack size timesRepeat: [logStream tab].
	logStream nextPutAll: 'openRecordGroup'.

	super openRecordGroup.! !

!LoggingUndoManager methodsFor: 'transaction' stamp: 'dg 7/6/2007 18:20'!
openRecordGroupNamed: aString

	super openRecordGroupNamed: aString.

	logStream nextPutAll: 'Named: '; print: aString.! !

!LoggingUndoManager methodsFor: 'undo/redo' stamp: 'dg 7/6/2007 18:16'!
redo
	| undoEntry |

	self redoHistory
		isEmpty ifTrue: [^self].

	undoEntry := self redoHistory removeLast.
	undoEntry redo.
	self undoHistory addLast: undoEntry.

	logStream cr.
	logStream nextPutAll: 'redone: '; print: undoEntry
! !

!LoggingUndoManager methodsFor: 'undo/redo' stamp: 'dg 7/6/2007 18:15'!
undo
	| undoEntry |

	self currentGroup ifNotNil: [^self].
		
	self undoHistory
		isEmpty ifTrue: [^self].

	undoEntry := self undoHistory removeLast.
	undoEntry undo.
	self redoHistory addLast: undoEntry.

	logStream cr.
	logStream nextPutAll: 'undone: '; print: undoEntry
! !

!UndoManager methodsFor: 'private' stamp: 'dg 7/5/2007 19:30'!
addRecord: undoRecord

	self currentGroup
		ifNil: [
			self historyInvalid
				ifTrue: ["now is the time to clean up invalid history caused by undoable entries"
					self clearUndoHistory.
					historyInvalid := false]
				ifFalse: [
					self undoHistory addLast: undoRecord.
					self keepHistoryInRange]]
		ifNotNil: [self currentGroup add: undoRecord]! !

!UndoManager methodsFor: 'accessing' stamp: 'dg 6/13/2007 10:09'!
clearRedoHistory

	redoHistory := OrderedCollection new! !

!UndoManager methodsFor: 'accessing' stamp: 'dg 7/5/2007 15:56'!
clearUndoHistory

	undoHistory := OrderedCollection new! !

!UndoManager methodsFor: 'transaction' stamp: 'swr 10/24/2007 11:42'!
closeRecordGroup
	"close current record group"
	| group |
	self groupStack notEmpty ifTrue: [
		group := self groupStack removeLast.
		group isEmpty ifFalse: [self addRecord: group]
	]! !

!UndoManager methodsFor: 'private' stamp: 'dg 6/4/2007 17:48'!
currentGroup
	^self groupStack notEmpty
		ifTrue: [self groupStack last]
		ifFalse: [nil]! !

!UndoManager methodsFor: 'transaction' stamp: 'dg 7/5/2007 19:19'!
doAndAddRecord: undoRecord

	| result |

	undoRecord undoable ifFalse:[(self confirm:'Operation: "',undoRecord description,'" is not undoable and destroys the undo history. Continue?') ifFalse:[^self]].

	self clearRedoHistory.
	
	result := undoRecord do.

	(undoRecord undoable)
		ifTrue: [
			self addRecord: undoRecord]
		ifFalse: [
			historyInvalid := true.
			self addRecord: undoRecord].

	^result! !

!UndoManager methodsFor: 'grouping' stamp: 'dg 6/11/2007 16:09'!
groupHistoryFrom: firstIndex to: lastIndex

	| newHistory newGroup |
	(self undoHistory size < 2) ifTrue: [^self].
	

	newHistory := OrderedCollection new.

	(firstIndex > 1) ifTrue: [
		(1 to: firstIndex - 1) do: [ :i | newHistory add: (undoHistory at: i)]
	].

	newGroup := UndoRecordGroup new.
	(firstIndex to: lastIndex) do: [ :i | newGroup add: (undoHistory at: i)].
	newHistory add: newGroup.

	(lastIndex < self undoHistory size) ifTrue: [
		(lastIndex + 1 to: self undoHistory size) do: [:i | newHistory add: (undoHistory at: i)]
	].

	undoHistory := newHistory.! !

!UndoManager methodsFor: 'private' stamp: 'dg 6/4/2007 17:48'!
groupStack
	^groupStack ifNil: [groupStack := OrderedCollection new]! !

!UndoManager methodsFor: 'accessing' stamp: 'dg 6/5/2007 15:51'!
historyInvalid
	^historyInvalid! !

!UndoManager methodsFor: 'initialize' stamp: 'dg 6/5/2007 16:16'!
initialize

	historyInvalid := false.
	maxHistoryLength := 0.! !

!UndoManager methodsFor: 'testing' stamp: 'JMM 12/5/2007 15:45'!
isUndoManagerDirty
	^self undoHistory isEmpty not! !

!UndoManager methodsFor: 'private' stamp: 'dg 6/5/2007 17:23'!
keepHistoryInRange

	(maxHistoryLength > 0 and: [self undoHistory size > maxHistoryLength]) ifTrue: [
		undoHistory removeFirst: (self undoHistory size - maxHistoryLength)
	]! !

!UndoManager methodsFor: 'accessing' stamp: 'dg 6/5/2007 16:12'!
maxHistoryLength
	^maxHistoryLength! !

!UndoManager methodsFor: 'accessing' stamp: 'dg 6/5/2007 16:13'!
maxHistoryLength: aNumber
	maxHistoryLength := aNumber! !

!UndoManager methodsFor: 'transaction' stamp: 'dg 6/14/2007 16:10'!
openRecordGroup
	"open a new recordGroup"
	self groupStack addLast: UndoRecordGroup new! !

!UndoManager methodsFor: 'transaction' stamp: 'dg 6/21/2007 15:42'!
openRecordGroupNamed: aString
	"open a new recordGroup and give it a description"
	self openRecordGroup.
	groupStack last description: aString.! !

!UndoManager methodsFor: 'undo/redo' stamp: 'dg 6/12/2007 17:42'!
redo
	| undoEntry |

	self redoHistory
		isEmpty ifTrue: [^self].

	undoEntry := self redoHistory removeLast.
	undoEntry redo.
	self undoHistory addLast: undoEntry! !

!UndoManager methodsFor: 'private' stamp: 'dg 6/4/2007 17:49'!
redoHistory
	^redoHistory ifNil: [redoHistory := OrderedCollection new]! !

!UndoManager methodsFor: 'initialize' stamp: 'dg 7/5/2007 19:24'!
reset

	self clearUndoHistory.
	self clearRedoHistory.
	groupStack := nil.! !

!UndoManager methodsFor: 'undo/redo' stamp: 'dg 6/5/2007 16:02'!
undo
	| undoEntry |

	self currentGroup ifNotNil: [^self].
		
	self undoHistory
		isEmpty ifTrue: [^self].

	undoEntry := self undoHistory removeLast.
	undoEntry undo.
	self redoHistory addLast: undoEntry! !

!UndoManager methodsFor: 'private' stamp: 'dg 6/4/2007 17:49'!
undoHistory
	^undoHistory ifNil: [undoHistory := OrderedCollection new]! !

TestCase subclass: #UndoManagerTest
	instanceVariableNames: 'undoManager'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Undo'!

!UndoManagerTest methodsFor: 'running' stamp: 'dg 6/11/2007 15:55'!
setUp

	undoManager := UndoManager new.! !

!UndoManagerTest methodsFor: 'testing' stamp: 'dg 6/14/2007 11:31'!
testGroupedUndo1
	"test of grouping history entrys afterwards by #groupHistoryFrom:to:"
	| aCollection |
	aCollection := OrderedCollection newFrom: {1. 2. 3}.

	undoManager doAndAddRecord: (
		UndoRecord
			do: (MessageSend receiver: aCollection selector: #removeLast )
			undo: (MessageSend receiver: aCollection selector: #addLast: argument: aCollection last)).

	undoManager doAndAddRecord: (
		UndoRecord
			do: (MessageSend receiver: aCollection selector: #removeLast )
			undo: (MessageSend receiver: aCollection selector: #addLast: argument: aCollection last)).

	undoManager groupHistoryFrom: 1 to: 2.
	undoManager undo.


	self assert: ((aCollection last = 3) and: [aCollection second = 2]).! !

!UndoManagerTest methodsFor: 'testing' stamp: 'dg 6/14/2007 16:27'!
testGroupedUndo2
	"test of grouping history entrys afterwards by #groupHistoryFrom:to:"
	| aCollection |
	aCollection := OrderedCollection newFrom: {1. 2. 3}.

	undoManager openRecordGroup.
	undoManager doAndAddRecord: (
		UndoRecord
			do: (MessageSend receiver: aCollection selector: #removeLast )
			undo: (MessageSend receiver: aCollection selector: #addLast: argument: aCollection last)).

	undoManager doAndAddRecord: (
		UndoRecord
			do: (MessageSend receiver: aCollection selector: #removeLast )
			undo: (MessageSend receiver: aCollection selector: #addLast: argument: aCollection last)).

	undoManager closeRecordGroup.

	"the undoManager should have grouped the two commands and placed this group in the undoHistory"
	self assert: ((undoManager undoHistory size = 1) and: [undoManager undoHistory last isKindOf: UndoRecordGroup]).

	undoManager undo.
	"both UndoRecords in the group should have been made undone"
	self assert: ((aCollection last = 3) and: [aCollection second = 2]).! !

!UndoManagerTest methodsFor: 'testing' stamp: 'dg 6/14/2007 16:28'!
testGrouping

	undoManager openRecordGroup.
	self assert: undoManager groupStack size = 1.
	undoManager closeRecordGroup.

	self assert: undoManager groupStack size = 0.
	self assert: undoManager undoHistory size = 1.! !

!UndoManagerTest methodsFor: 'testing' stamp: 'dg 6/11/2007 16:10'!
testMaxHistoryLength

	| aCollection |
	aCollection := OrderedCollection newFrom: {1. 2. 3}.
	undoManager maxHistoryLength: 5.

	10 timesRepeat: [
		undoManager doAndAddRecord: (
			UndoRecord
				do: (MessageSend receiver: aCollection selector: #size)
				undo: (MessageSend receiver: aCollection selector: #size)).
	].

	self assert: undoManager undoHistory size = 5.! !

!UndoManagerTest methodsFor: 'testing' stamp: 'dg 6/11/2007 16:10'!
testUndo1

	| aCollection |
	aCollection := OrderedCollection new.

	undoManager doAndAddRecord: (
		UndoRecord
			do: (MessageSend receiver: aCollection selector: #addLast: argument: 1)
			undo: (MessageSend receiver: aCollection selector: #removeLast)).

	undoManager undo.

	self assert: aCollection size = 0.! !

!UndoManagerTest methodsFor: 'testing' stamp: 'dg 6/11/2007 16:10'!
testUndo2

	| aCollection |
	aCollection := OrderedCollection newFrom: {1. 2. 3}.

	undoManager doAndAddRecord: (
		UndoRecord
			do: (MessageSend receiver: aCollection selector: #removeLast )
			undo: (MessageSend receiver: aCollection selector: #addLast: argument: aCollection last)).

	undoManager undo.

	self assert: aCollection last = 3.! !
